/* clasha.h - classes for pure C by Shaos

 Copyright (c) 2007,2014 A.A.Shabarshin <me@shaos.net>

 Permission is hereby granted, free of charge, to any person obtaining
 a copy of this software and associated documentation files (the "Software"),
 to deal in the Software without restriction, including without limitation
 the rights to use, copy, modify, merge, publish, distribute, sublicense,
 and/or sell copies of the Software, and to permit persons to whom
 the Software is furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included
 in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.

*/

#ifndef IMPLEMENTATION
#ifdef CLASS
#undef CLASS
#endif
#endif

#ifndef CLASS

#ifndef CLASHA_H
#define CLASHA_H
#include <stdarg.h>
#define CLASHA_VER_STR "v0.1"
#define CLASHA_VER 0x0001 /* (major<<8)|minor */
#define Q(x) #x
#define QQ(x) Q(x)
#define $(x,y) $$(x,_,y)
#define $$(x,y,z) x##y##z
#define NEW(x) $(new,x)
#define DEL(x) $(del,x)
#define BAD(x) $(bad,x)
#define SET(x) $(set,x)
#define CON(x) $(x,con)
#define CONS(x) $(x,cons)
#define DES(x) $(x,des)
#define HASH(x) $(x,hash)
#ifdef LIGHT
#define CLASS_BEGIN(x) typedef struct $(t,x) {
#else
#define CLASS_BEGIN(x) typedef struct $(t,x) { int _h_,_n_;
#ifdef HEAVY
#ifdef __cplusplus
extern "C" {
#endif
/* optional class registration for delete method */
void register_class(int h,void(*f)(void*));
/* optional common delete that will call proper method */
void common_delete(void *p);
#ifdef __cplusplus
}
#endif
#define delete(x) common_delete(x)
#endif
#endif
#define CLASS_END(x) } x;
#ifdef __cplusplus
#define PUBLIC extern "C"
#else
#define PUBLIC extern
#endif
#define METHOD(x) $(CLASS,x)
#define STATIC(x) $(CLASS,x)
#define VIRTUAL(x) (*x)
#define BEGIN CLASS_BEGIN(CLASS)
#define END CLASS_END(CLASS) \
 PUBLIC void CON(CLASS) (CLASS*,int,...); \
 PUBLIC void CONS(CLASS) (CLASS*,int,va_list); \
 PUBLIC void DES(CLASS) (CLASS*); \
 PUBLIC CLASS* NEW(CLASS) (int,...); \
 PUBLIC void DEL(CLASS) (void*); \
 PUBLIC int BAD(CLASS) (CLASS*);
#endif

#else

#include <stdlib.h>

/* private methods */

#ifndef LIGHT
int HASH(CLASS) = 0;

void SET(CLASS) (CLASS* p)
{
 int c,h=0;
 char *o,s[] = QQ(CLASS);
 if(!HASH(CLASS))
 {
   o = s;
   while((c=*o++)) h += (h<<5) + c;
   HASH(CLASS) = h;
 }
 p->_h_ = HASH(CLASS);
}
#endif

/* public methods */

int BAD(CLASS) (CLASS* p)
{
#ifndef LIGHT
 if(HASH(CLASS)!=p->_h_) return -1;
#endif
 return 0;
}

void CON(CLASS) (CLASS *p, int t, ...)
{
 va_list a;
#ifndef LIGHT
 SET(CLASS) (p);
#endif
 va_start(a,t);
 CONS(CLASS) (p,t,a);
 va_end(a);
}

CLASS* NEW(CLASS) (int t, ...)
{
 int i;
 CLASS* p;
 p = (CLASS*)malloc(sizeof(CLASS)*((t>0)?(size_t)t:1));
 if(p)
 {
  if(t > 0) /* array */
  {
    for(i=0;i<t;i++)
    {
      CON(CLASS) (&p[i],0);
#ifndef LIGHT
      p[i]._n_ = t - i;
#endif
    }
  }
  else
  {
    va_list a;
    va_start(a,t);
#ifndef LIGHT
    SET(CLASS) (p);
#endif
    CONS(CLASS) (p,t,a);
    va_end(a);
#ifndef LIGHT
    p->_n_ = 1;
#endif
  }
#ifndef LIGHT
#ifdef HEAVY
  register_class(p->_h_,DEL(CLASS));
#endif
#endif
 }
 return p;
}

void DEL(CLASS) (void* v)
{
 CLASS* p = (CLASS*)v;
#ifndef LIGHT
 int i,n = p->_n_;
 if(BAD(CLASS)(p))
 {
    exit(-1);
 }
#endif
 DES(CLASS) (p);
#ifndef LIGHT
 p->_h_ = 0;
 for(i=1;i<n;i++)
 {
   DES(CLASS) (&p[i]);
   p[i]._h_ = 0;
 }
#endif
 free(p);
}

#undef CLASS

#endif
